/*
 * Fake TTY char devices, giving access to raw input device events.
 *
 * Copyright (c) 2005 Aivils Stoss
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#define FTTY_MINOR_BASE		96
#define FTTY_MINORS		32
#define FTTY_BUFFER_SIZE	64

#include <linux/version.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/input.h>
#include <linux/major.h>
#include <linux/smp_lock.h>
#include <linux/device.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
#include <linux/devfs_fs_kernel.h>
#endif

/*#include <linux/kbd_kern.h>*/
#include <linux/interrupt.h>
#include <linux/kd.h>
#include <linux/vt.h>

#include "version.h"

struct kbd_struct {

	unsigned char lockstate;
/* 8 modifiers - the names do not have any meaning at all;
   they can be associated to arbitrarily chosen keys */
#define VC_SHIFTLOCK    KG_SHIFT        /* shift lock mode */
#define VC_ALTGRLOCK    KG_ALTGR        /* altgr lock mode */
#define VC_CTRLLOCK     KG_CTRL         /* control lock mode */
#define VC_ALTLOCK      KG_ALT          /* alt lock mode */
#define VC_SHIFTLLOCK   KG_SHIFTL       /* shiftl lock mode */
#define VC_SHIFTRLOCK   KG_SHIFTR       /* shiftr lock mode */
#define VC_CTRLLLOCK    KG_CTRLL        /* ctrll lock mode */
#define VC_CTRLRLOCK    KG_CTRLR        /* ctrlr lock mode */
	unsigned char slockstate;       /* for `sticky' Shift, Ctrl, etc. */

	unsigned char ledmode:2;        /* one 2-bit value */
#define LED_SHOW_FLAGS 0        /* traditional state */
#define LED_SHOW_IOCTL 1        /* only change leds upon ioctl */
#define LED_SHOW_MEM 2          /* `heartbeat': peek into memory */

	unsigned char ledflagstate:4;   /* flags, not lights */
	unsigned char default_ledflagstate:4;
#define VC_SCROLLOCK    0       /* scroll-lock mode */
#define VC_NUMLOCK      1       /* numeric lock mode */
#define VC_CAPSLOCK     2       /* capslock mode */
#define VC_KANALOCK     3       /* kanalock mode */

	unsigned char kbdmode:2;        /* one 2-bit value */
#define VC_XLATE        0       /* translate keycodes using keymap */
#define VC_MEDIUMRAW    1       /* medium raw (keycode) mode */
#define VC_RAW          2       /* raw (scancode) mode */
#define VC_UNICODE      3       /* Unicode mode */

	unsigned char modeflags:5;
#define VC_APPLIC       0       /* application key mode */
#define VC_CKMODE       1       /* cursor key mode */
#define VC_REPEAT       2       /* keyboard repeat */
#define VC_CRLF         3       /* 0 - enter sends CR, 1 - enter sends CRLF */
#define VC_META         4       /* 0 - meta, 1 - meta=prefix with ESC */
};

struct evdev {
	int exist;
	int open;
	int minor;
	char name[16];
	struct input_handle handle;
	wait_queue_head_t wait;
	struct ftty_list *grab;
	struct list_head list;
	struct kbd_struct kbd_table;
	struct kbd_repeat kbrep;
	unsigned char ledstate;
	unsigned char ledioctl;
};

#define FTTY_FLIPBUF_SIZE 512

struct ftty_flip_buffer {
	char            *char_buf_ptr;
	unsigned char   *flag_buf_ptr;
	int             count;
	unsigned char   char_buf[FTTY_FLIPBUF_SIZE];
	char            flag_buf[FTTY_FLIPBUF_SIZE];
	int             tail;
};

struct ftty_list {
	struct ftty_flip_buffer flip;
	struct fasync_struct *fasync;
	struct evdev *evdev;
	struct list_head node;
	spinlock_t read_lock;
};

static struct ledptr {
	unsigned int *addr;
	unsigned int mask;
	unsigned char valid:1;
} ledptrs[3];

static struct evdev *evdev_table[FTTY_MINORS];
static int sysrq_alt;
static void ftty_bh(unsigned long dummy);
static struct input_handler ftty_handler;

DECLARE_TASKLET_DISABLED(ftty_tasklet, ftty_bh, 0);

static inline void ftty_set_leds(void)
{
	tasklet_schedule(&ftty_tasklet);
}

static inline void ftty_insert_flip_char(struct ftty_list *list,
	   unsigned char ch, char flag)
{
	if (list->flip.count < FTTY_FLIPBUF_SIZE) {
		list->flip.count++;
		*list->flip.flag_buf_ptr++ = flag;
		*list->flip.char_buf_ptr++ = ch;
	}
}

static void put_queue(struct ftty_list *list, int ch)
{
	ftty_insert_flip_char(list, ch, 0);
}

#if defined(CONFIG_X86) || defined(CONFIG_IA64) || defined(CONFIG_ALPHA) ||\
    defined(CONFIG_MIPS) || defined(CONFIG_PPC) || defined(CONFIG_SPARC32) ||\
    defined(CONFIG_SPARC64) || defined(CONFIG_PARISC) || defined(CONFIG_SUPERH) ||\
    (defined(CONFIG_ARM) && defined(CONFIG_KEYBOARD_ATKBD) && !defined(CONFIG_ARCH_RPC))

#define HW_RAW(dev) (test_bit(EV_MSC, dev->evbit) && test_bit(MSC_RAW, dev->mscbit) &&\
			((dev)->id.bustype == BUS_I8042) && ((dev)->id.vendor == 0x0001) && ((dev)->id.product == 0x0001))

static unsigned short x86_keycodes[256] =
	{ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
	 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
	 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
	 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79,
	 80, 81, 82, 83, 84,118, 86, 87, 88,115,120,119,121,112,123, 92,
	284,285,309,  0,312, 91,327,328,329,331,333,335,336,337,338,339,
	367,288,302,304,350, 89,334,326,267,126,268,269,125,347,348,349,
	360,261,262,263,268,376,100,101,321,316,373,286,289,102,351,355,
	103,104,105,275,287,279,306,106,274,107,294,364,358,363,362,361,
	291,108,381,281,290,272,292,305,280, 99,112,257,258,359,113,114,
	264,117,271,374,379,265,266, 93, 94, 95, 85,259,375,260, 90,116,
	377,109,111,277,278,282,283,295,296,297,299,300,301,293,303,307,
	308,310,313,314,315,317,318,319,320,357,322,323,324,325,276,330,
	332,340,365,342,343,344,345,346,356,270,341,368,369,370,371,372 };

static int emulate_raw(struct ftty_list *list, unsigned int keycode,
		       unsigned char up_flag)
{
	int code;

	switch (keycode) {
		case KEY_PAUSE:
			put_queue(list, 0xe1);
			put_queue(list, 0x1d | up_flag);
			put_queue(list, 0x45 | up_flag);
			break;

		case KEY_HANGUEL:
			if (!up_flag)
				put_queue(list, 0xf2);
			break;

		case KEY_HANJA:
			if (!up_flag)
				put_queue(list, 0xf1);
			break;

		case KEY_SYSRQ:
			/*
			 * Real AT keyboards (that's what we're trying
			 * to emulate here emit 0xe0 0x2a 0xe0 0x37 when
			 * pressing PrtSc/SysRq alone, but simply 0x54
			 * when pressing Alt+PrtSc/SysRq.
			 */
			if (sysrq_alt) {
				put_queue(list, 0x54 | up_flag);
			} else {
				put_queue(list, 0xe0);
				put_queue(list, 0x2a | up_flag);
				put_queue(list, 0xe0);
				put_queue(list, 0x37 | up_flag);
			}
			break;

		default:
			if (keycode > 255)
				return -1;

			code = x86_keycodes[keycode];
			if (!code)
				return -1;

			if (code & 0x100)
				put_queue(list, 0xe0);
			put_queue(list, (code & 0x7f) | up_flag);

			break;
	}

	return 0;
}

#else

#define HW_RAW(dev)	0

#warning "Cannot generate rawmode keyboard for your architecture yet."

static int emulate_raw(struct ftty_list *list, unsigned int keycode, unsigned char up_flag)
{
	if (keycode > 127)
		return -1;

	put_queue(list, keycode | up_flag);
	return 0;
}
#endif

static void ftty_keycode(struct ftty_list *list,
			unsigned int keycode, int down,
			int hw_raw)
{
	struct kbd_struct *kbd;

	kbd = &list->evdev->kbd_table;

	/*Only VC_RAW VC_MEDIUMRAW */
	if (kbd->kbdmode == VC_RAW && !hw_raw)
		if (emulate_raw(list, keycode, !down << 7))
			if (keycode < BTN_MISC)
				printk(KERN_WARNING "faketty: can't emulate rawmode for keycode %d\n", keycode);

	if (kbd->kbdmode == VC_MEDIUMRAW) {
		/*
		 * This is extended medium raw mode, with keys above 127
		 * encoded as 0, high 7 bits, low 7 bits, with the 0 bearing
		 * the 'up' flag if needed. 0 is reserved, so this shouldn't
		 * interfere with anything else. The two bytes after 0 will
		 * always have the up flag set not to interfere with older
		 * applications. This allows for 16384 different keycodes,
		 * which should be enough.
		 */
		if (keycode < 128) {
			put_queue(list, keycode | (!down << 7));
		} else {
			put_queue(list, !down << 7);
			put_queue(list, (keycode >> 7) | 0x80);
			put_queue(list, keycode | 0x80);
		}
	}
}

static void ftty_rawcode(struct ftty_list *list, unsigned char data)
{
	if (list->evdev->kbd_table.kbdmode == VC_RAW)
		put_queue(list, data);
}

static void ftty_event(struct input_handle *handle, unsigned int type, unsigned int code, int value)
{
	struct evdev *evdev = handle->private;
	struct ftty_list *list;

	if (evdev->grab) {
		list = evdev->grab;

		if (type == EV_MSC && code == MSC_RAW && HW_RAW(handle->dev))
			ftty_rawcode(list, value);
		if (type == EV_KEY)
			ftty_keycode(list, code, value, HW_RAW(handle->dev));
		kill_fasync(&list->fasync, SIGIO, POLL_IN);
	}
	/* Multiple ftty device read isn't supported */
	/* Device grab is default */
	wake_up_interruptible(&evdev->wait);
}

static int ftty_fasync(int fd, struct file *file, int on)
{
	int retval;
	struct ftty_list *list = file->private_data;
	retval = fasync_helper(fd, file, on, &list->fasync);
	return retval < 0 ? retval : 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
static int ftty_flush(struct file * file, fl_owner_t id)
#else
static int ftty_flush(struct file * file)
#endif
{
	struct ftty_list *list = file->private_data;
	if (!list->evdev->exist) return -ENODEV;
	return input_flush_device(&list->evdev->handle, file);
}

static void ftty_free(struct evdev *evdev)
{
	evdev_table[evdev->minor] = NULL;
	kfree(evdev);
}

static int ftty_release(struct inode * inode, struct file * file)
{
	struct ftty_list *list = file->private_data;

	if (list->evdev->grab == list) {
		input_release_device(&list->evdev->handle);
		list->evdev->grab = NULL;
	}

	ftty_fasync(-1, file, 0);
	list_del(&list->node);

	if (!--list->evdev->open) {
		if (list->evdev->exist)
			input_close_device(&list->evdev->handle);
		else
			ftty_free(list->evdev);
	}

	kfree(list);
	return 0;
}

static int ftty_open(struct inode * inode, struct file * file)
{
	struct ftty_list *list;
	int i = iminor(inode) - FTTY_MINOR_BASE;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
	int accept_err;
#endif

	if (i >= FTTY_MINORS || !evdev_table[i] || !evdev_table[i]->exist)
		return -ENODEV;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
	if ((accept_err = input_accept_process(&(evdev_table[i]->handle), file)))
		return accept_err;
#endif

	if (!(list = kmalloc(sizeof(struct ftty_list), GFP_KERNEL)))
		return -ENOMEM;
	memset(list, 0, sizeof(struct ftty_list));

	spin_lock_init(&list->read_lock);
	list->evdev = evdev_table[i];
	list_add_tail(&list->node, &evdev_table[i]->list);
	file->private_data = list;
	list->flip.char_buf_ptr = list->flip.char_buf;
	list->flip.flag_buf_ptr = list->flip.flag_buf;

	if (!list->evdev->open++)
		if (list->evdev->exist)
			input_open_device(&list->evdev->handle);

	list->evdev->kbd_table.kbdmode = VC_RAW;
	list->evdev->kbd_table.ledmode = LED_SHOW_FLAGS;
	list->evdev->ledstate = 0xff;

	if (!input_grab_device(&list->evdev->handle)) {
		list->evdev->grab = list;
	}
	else
		printk(KERN_INFO "faketty: cannot grab device ftty%d\n", i);

	return 0;
}

static ssize_t ftty_write(struct file * file, const char __user * buffer, size_t count, loff_t *ppos)
{
	struct ftty_list *list = file->private_data;

	if (!list->evdev->exist) return -ENODEV;

	return count;
}

static ssize_t ftty_read(struct file * file, char __user * buffer, size_t count, loff_t *ppos)
{
	struct ftty_list *list = file->private_data;
	int retval;
	unsigned long 	flags;

	if (list->flip.count == list->flip.tail && list->evdev->exist && (file->f_flags & O_NONBLOCK))
		return -EAGAIN;

	retval = wait_event_interruptible(list->evdev->wait,
		list->flip.count != list->flip.tail || (!list->evdev->exist));

	if (retval)
		return retval;

	if (!list->evdev->exist)
		return -ENODEV;

	while (list->flip.count != list->flip.tail && retval + 1 <= count &&
	       list->flip.tail < FTTY_BUFFER_SIZE) {
		if (copy_to_user(buffer + retval, list->flip.char_buf + list->flip.tail,1))
			return -EFAULT;
		list->flip.tail = list->flip.tail + 1;
		retval++;
	}

	if (list->flip.count == list->flip.tail) {
		spin_lock_irqsave(&list->read_lock, flags);
		list->flip.char_buf_ptr = list->flip.char_buf;
		list->flip.flag_buf_ptr = list->flip.flag_buf;
		list->flip.count = 0;
		list->flip.tail = 0;
		spin_unlock_irqrestore(&list->read_lock, flags);
	}

	return retval;
}

/* No kernel lock - fine */
static unsigned int ftty_poll(struct file *file, poll_table *wait)
{
	struct ftty_list *list = file->private_data;
	poll_wait(file, &list->evdev->wait, wait);
	if (list->flip.count != list->flip.tail)
		return POLLIN | POLLRDNORM;
	return 0;
}

/*
 * Making beeps and bells. 
 */
static void kd_nosound(unsigned long ignored)
{
	struct list_head * node;

	list_for_each(node,&ftty_handler.h_list) {
		struct input_handle *handle = to_handle_h(node);
		if (test_bit(EV_SND, handle->dev->evbit)) {
			if (test_bit(SND_TONE, handle->dev->sndbit))
				input_event(handle->dev, EV_SND, SND_TONE, 0);
			if (test_bit(SND_BELL, handle->dev->sndbit))
				input_event(handle->dev, EV_SND, SND_BELL, 0);
		}
	}
}

static struct timer_list kd_mksound_timer =
		TIMER_INITIALIZER(kd_nosound, 0, 0);

void kd_mksound(unsigned int hz, unsigned int ticks)
{
	struct list_head * node;

	del_timer(&kd_mksound_timer);

	if (hz) {
		list_for_each_prev(node,&ftty_handler.h_list) {
			struct input_handle *handle = to_handle_h(node);
			if (test_bit(EV_SND, handle->dev->evbit)) {
				if (test_bit(SND_TONE, handle->dev->sndbit)) {
					input_event(handle->dev, EV_SND, SND_TONE, hz);
					break;
				}
				if (test_bit(SND_BELL, handle->dev->sndbit)) {
					input_event(handle->dev, EV_SND, SND_BELL, 1);
					break;
				}
			}
		}
		if (ticks)
			mod_timer(&kd_mksound_timer, jiffies + ticks);
	} else
		kd_nosound(0);
}

/*
 * Setting the keyboard rate.
 */

int kbd_rate(struct input_handle *handle, struct kbd_repeat *rep)
{
	struct input_dev *dev = handle->dev;
	unsigned int d = 0;
	unsigned int p = 0;

	if (test_bit(EV_REP, dev->evbit)) {
		if (rep->delay > 0)
			input_event(dev, EV_REP, REP_DELAY, rep->delay);
		if (rep->period > 0)
			input_event(dev, EV_REP, REP_PERIOD, rep->period);
		d = dev->rep[REP_DELAY];
		p = dev->rep[REP_PERIOD];
	}
	rep->delay  = d;
	rep->period = p;
	return 0;
}

/*
 * The leds display either (i) the status of NumLock, CapsLock, ScrollLock,
 * or (ii) whatever pattern of lights people want to show using KDSETLED,
 * or (iii) specified bits of specified words in kernel memory.
 */
unsigned char ftty_getledstate(struct evdev *evdev)
{
	return evdev->ledstate;
}

void ftty_setledstate(struct evdev *evdev, unsigned int led)
{
	if (!(led & ~7)) {
		evdev->ledioctl = led;
		evdev->kbd_table.ledmode = LED_SHOW_IOCTL;
	} else {
		evdev->kbd_table.ledmode = LED_SHOW_FLAGS;
	}
	ftty_set_leds();
}

static inline unsigned char ftty_getleds(struct evdev *evdev)
{
	struct kbd_struct *kbd = &evdev->kbd_table;
	unsigned char leds;
	int i;

	if (kbd->ledmode == LED_SHOW_IOCTL)
		return evdev->ledioctl;

	leds = kbd->ledflagstate;

	if (kbd->ledmode == LED_SHOW_MEM) {
		for (i = 0; i < 3; i++)
			if (ledptrs[i].valid) {
				if (*ledptrs[i].addr & ledptrs[i].mask)
					leds |= (1 << i);
				else
					leds &= ~(1 << i);
			}
	}
	return leds;
}


int ftty_ioctl(struct inode *inode, struct file * file,
	     unsigned int cmd, unsigned long arg)
{
	struct ftty_list *list = file->private_data;
	struct evdev *evdev = list->evdev;
 
	struct input_handle *handle = &evdev->handle;
	struct kbd_struct * kbd;
	unsigned char ucval;
	void __user *up = (void __user *)arg;
	int perm;
	unsigned int ignore = cmd >> 8;
	
	if (!evdev->exist) return -ENODEV;

	/*
	 * To have permissions to do most of the vt ioctls, we either have
	 * to be the owner of the tty, or have CAP_SYS_TTY_CONFIG.
	 */
	perm = 0;
	if (capable(CAP_SYS_TTY_CONFIG))
		perm = 1;

	kbd = &list->evdev->kbd_table;
	/*printk("cmd=%04x arg=%08lx\n", cmd, arg);*/
	if(ignore == 0x54) /* T terminal ioctl's ingnore */
		return 0;
	if(ignore == 0x56) /* V VT ioctl's ingnore */
		return 0;
	switch (cmd) {
	case KIOCSOUND:
		if (!perm)
			return -EPERM;
		if (arg)
			arg = CLOCK_TICK_RATE / arg;
		kd_mksound(arg, 0);
		return 0;

	case KDMKTONE:
		if (!perm)
			return -EPERM;
	{
		unsigned int ticks, count;
		
		/*
		 * Generate the tone for the appropriate number of ticks.
		 * If the time is zero, turn off sound ourselves.
		 */
		ticks = HZ * ((arg >> 16) & 0xffff) / 1000;
		count = ticks ? (arg & 0xffff) : 0;
		if (count)
			count = CLOCK_TICK_RATE / count;
		kd_mksound(count, ticks);
		return 0;
	}
	case KDGKBTYPE:
		/*
		 * this is naive.
		 */
		ucval = KB_101;
		goto setchar;

		/*
		 * These cannot be implemented on any machine that implements
		 * ioperm() in user level (such as Alpha PCs) or not at all.
		 *
		 * XXX: you should never use these, just call ioperm directly..
		 */
#ifdef CONFIG_X86
	case KDADDIO:
	case KDDELIO:
	case KDENABIO:
	case KDDISABIO:
		return 0;
#endif

	/* Linux m68k/i386 interface for setting the keyboard delay/repeat rate */
		
	case KDKBDREP:
	{
		struct kbd_repeat * kbrep = &list->evdev->kbrep;
		int err;
		
		if (!capable(CAP_SYS_TTY_CONFIG))
			return -EPERM;

		if (copy_from_user(kbrep, up, sizeof(struct kbd_repeat)))
			return -EFAULT;
		err = kbd_rate(handle, kbrep);
		if (err)
			return err;
		if (copy_to_user(up, kbrep, sizeof(struct kbd_repeat)))
			return -EFAULT;
		return 0;
	}

	case KDSETMODE:
		return 0;

	case KDGETMODE:
		ucval = KD_GRAPHICS;
		goto setint;

	case KDMAPDISP:
	case KDUNMAPDISP:
		return -EINVAL;

	case KDSKBMODE:
		if (!perm)
			return -EPERM;
		switch(arg) {
		  case K_RAW:
			kbd->kbdmode = VC_RAW;
			break;
		  case K_MEDIUMRAW:
			kbd->kbdmode = VC_MEDIUMRAW;
			break;
		  case K_XLATE:
			kbd->kbdmode = VC_XLATE;
			break;
		  case K_UNICODE:
			kbd->kbdmode = VC_UNICODE;
			break;
		  default:
			return -EINVAL;
		}
		return 0;

	case KDGKBMODE:
		ucval = ((kbd->kbdmode == VC_RAW) ? K_RAW :
				 (kbd->kbdmode == VC_MEDIUMRAW) ? K_MEDIUMRAW :
				 (kbd->kbdmode == VC_UNICODE) ? K_UNICODE :
				 K_XLATE);
		goto setint;

	/* this could be folded into KDSKBMODE, but for compatibility
	   reasons it is not so easy to fold KDGKBMETA into KDGKBMODE */
	case KDSKBMETA:
	case KDGKBMETA:
		ucval = 0;
	setint:
		return put_user(ucval, (int __user *)arg); 

	case KDGETKEYCODE:
	case KDSETKEYCODE:
	case KDGKBENT:
	case KDSKBENT:
	case KDGKBSENT:
	case KDSKBSENT:
	case KDGKBDIACR:
	case KDSKBDIACR:
		return 0;

	/* the ioctls below read/set the flags usually shown in the leds */
	/* don't use them - they will go away without warning */
	case KDGKBLED:
		ucval = kbd->ledflagstate | (kbd->default_ledflagstate << 4);
		goto setchar;

	case KDSKBLED:
		if (!perm)
			return -EPERM;
		if (arg & ~0x77)
			return -EINVAL;
		kbd->ledflagstate = (arg & 7);
		kbd->default_ledflagstate = ((arg >> 4) & 7);
    		ftty_set_leds();
		return 0;

	/* the ioctls below only set the lights, not the functions */
	/* for those, see KDGKBLED and KDSKBLED above */
	case KDGETLED:
		ucval = ftty_getledstate(list->evdev);
	setchar:
		return put_user(ucval, (char __user *)arg);

	case KDSETLED:
		if (!perm)
		  return -EPERM;
		ftty_setledstate(list->evdev, arg);
		return 0;

	case KDSIGACCEPT:
	case PIO_FONT:
	case GIO_FONT:
	case PIO_CMAP:
	case GIO_CMAP:
	case PIO_FONTX:
	case GIO_FONTX:
	case PIO_FONTRESET:
	case KDFONTOP:
	case PIO_SCRNMAP:
	case GIO_SCRNMAP:
	case PIO_UNISCRNMAP:
	case GIO_UNISCRNMAP:
	case PIO_UNIMAPCLR:
	case PIO_UNIMAP:
	case GIO_UNIMAP:
	 return 0;
	default:
		printk(KERN_INFO "faketty: forgotten ioctl cmd=%x arg=%lx\n", cmd, arg);
	 return 0;
/* 		return -ENOIOCTLCMD; */
	}
}


static struct file_operations ftty_fops = {
	.owner =	THIS_MODULE,
	.read =		ftty_read,
	.write =	ftty_write,
	.poll =		ftty_poll,
	.open =		ftty_open,
	.release =	ftty_release,
	.ioctl =	ftty_ioctl,
	.fasync =	ftty_fasync,
	.flush =	ftty_flush
};

static struct input_handle *ftty_connect(struct input_handler *handler, struct input_dev *dev, const struct input_device_id *id)
{
	struct evdev *evdev;
	int minor, i;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
	struct class_device *cdev;
	int ret;
#endif

	for (i = KEY_RESERVED; i < BTN_MISC; i++)
		if (test_bit(i, dev->keybit)) break;

	if ((i == BTN_MISC) && !test_bit(EV_SND, dev->evbit)) 
		return NULL;

	for (minor = 0; minor < FTTY_MINORS && evdev_table[minor]; minor++)
	{
		if (!evdev_table[minor]->exist && evdev_table[minor]->open) /* reuse */
		{
			evdev = evdev_table[minor];
			evdev->handle.dev = dev;
			evdev->handle.handler = handler;
			evdev->handle.private = evdev; /* should already be */
			evdev->exist = 1;
			printk(KERN_INFO "faketty: reusing ftty%d for vendor/product 0x%x/0x%x\n", minor, dev->id.vendor, dev->id.product);
			input_open_device(&evdev->handle);
			wake_up_interruptible(&evdev->wait);
			return &evdev->handle;
		}
	}
	if (minor == FTTY_MINORS) {
		printk(KERN_ERR "faketty: no more free ftty devices\n");
		return NULL;
	}

	if (!(evdev = kmalloc(sizeof(struct evdev), GFP_KERNEL)))
		return NULL;
	memset(evdev, 0, sizeof(struct evdev));

	INIT_LIST_HEAD(&evdev->list);
	init_waitqueue_head(&evdev->wait);

	evdev->exist = 1;
	evdev->minor = minor;
	evdev->handle.dev = dev;
	evdev->handle.name = evdev->name;
	evdev->handle.handler = handler;
	evdev->handle.private = evdev;
	sprintf(evdev->name, "ftty%d", minor);

	evdev_table[minor] = evdev;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
	devfs_mk_cdev(MKDEV(INPUT_MAJOR, FTTY_MINOR_BASE + minor),
			S_IFCHR|S_IRUGO|S_IWUSR, "input/ftty%d", minor);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
	cdev = class_device_create(&input_class, &dev->cdev,
			MKDEV(INPUT_MAJOR, FTTY_MINOR_BASE + minor),
			dev->cdev.dev, evdev->name);

	ret = sysfs_create_link(&input_class.subsys.kset.kobj, &cdev->kobj,
			  evdev->name);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
	class_device_create(&input_class, NULL,
				MKDEV(INPUT_MAJOR, FTTY_MINOR_BASE + minor),
				dev->dev, "ftty%d", minor);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
	class_device_create(input_class,
				MKDEV(INPUT_MAJOR, FTTY_MINOR_BASE + minor),
				dev->dev, "ftty%d", minor);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,2)
	class_simple_device_add(input_class,
				MKDEV(INPUT_MAJOR, FTTY_MINOR_BASE + minor),
				dev->dev, "ftty%d", minor);
#endif

	return &evdev->handle;
}

static void ftty_disconnect(struct input_handle *handle)
{
	struct evdev *evdev = handle->private;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
	class_device_destroy(&input_class,
				MKDEV(INPUT_MAJOR, FTTY_MINOR_BASE + evdev->minor));
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13)
	class_device_destroy(input_class,
				MKDEV(INPUT_MAJOR, FTTY_MINOR_BASE + evdev->minor));
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,2)
	class_simple_device_remove(MKDEV(INPUT_MAJOR, FTTY_MINOR_BASE + evdev->minor));
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
	devfs_remove("input/ftty%d", evdev->minor);
#endif

	evdev->exist = 0;

	if (evdev->open) {
		input_close_device(handle);
		wake_up_interruptible(&evdev->wait);
	} else
		ftty_free(evdev);
}

static struct input_device_id ftty_ids[] = {
	{
	.flags = INPUT_DEVICE_ID_MATCH_EVBIT,
	.evbit = { BIT(EV_KEY) },
	},
	
	{
	.flags = INPUT_DEVICE_ID_MATCH_EVBIT,
	.evbit = { BIT(EV_SND) },
	},

	{ },    /* Terminating entry */
};

MODULE_DEVICE_TABLE(input, ftty_ids);

static struct input_handler ftty_handler = {
	.event =	ftty_event,
	.connect =	ftty_connect,
	.disconnect =	ftty_disconnect,
	.fops =		&ftty_fops,
	.minor =	FTTY_MINOR_BASE,
	.name =		"ftty",
	.id_table =	ftty_ids,
};

static void ftty_bh(unsigned long dummy)
{
	struct list_head * node;
	unsigned char leds;

	list_for_each(node,&ftty_handler.h_list) {
		struct input_handle * handle = to_handle_h(node);
		struct evdev *evdev = handle->private;
		
		leds = ftty_getleds(evdev);
		if (leds != evdev->ledstate) {
			input_event(handle->dev, EV_LED, LED_SCROLLL, !!(leds & 0x01));
			input_event(handle->dev, EV_LED, LED_NUML,    !!(leds & 0x02));
			input_event(handle->dev, EV_LED, LED_CAPSL,   !!(leds & 0x04));
			input_sync(handle->dev);
			evdev->ledstate = leds;
		}
	}
}

static int __init ftty_init(void)
{
	printk(KERN_INFO "faketty: version %s\n", FTTY_VERSION);
	input_register_handler(&ftty_handler);
	tasklet_enable(&ftty_tasklet);
	tasklet_schedule(&ftty_tasklet);

	return 0;
}

static void __exit ftty_exit(void)
{
	tasklet_kill(&ftty_tasklet);
	tasklet_disable(&ftty_tasklet);
	input_unregister_handler(&ftty_handler);
}

module_init(ftty_init);
module_exit(ftty_exit);

MODULE_AUTHOR("Aivils Stoss");
MODULE_DESCRIPTION("Fake TTY input driver");
MODULE_LICENSE("GPL");
